#include <stdio.h>
#include <stdlib.h>
#include <htc.h>
#include <math.h>

// Set crystal frequency for configuring delays
#ifndef _XTAL_FREQ
#define _XTAL_FREQ 8000000 // Internal oscillator will be set to 8MHz
#endif

// Give outputs meaningful names
#define seg_a RA0
#define seg_b RA1
#define seg_c RB6
#define seg_d RB5
#define seg_e RB4
#define seg_f RA6
#define seg_g RA7
#define seg_dp RB7
#define select_7seg RB2 // Use this output to MUX between 7-segment displays

// Handy look-up table for decimal numbers 0,1,...,9 on 7-segment display.
// Each byte value is interpreted as [dp, a, b, ..., g].
const unsigned char LUT_7SEG[10] = {126,48,109,121,51,91,31,112,127,115};
const unsigned char DP_MASK = 128;

// Set configuration
__CONFIG(FOSC_INTOSCIO & WDTE_OFF & PWRTE_OFF & MCLRE_OFF
         & BOREN_OFF & LVP_ON & CPD_OFF & WRT_OFF & DEBUG_OFF & CP_OFF);
		 
__CONFIG(FCMEN_OFF & IESO_OFF);

// Initialisation function
void init(void)
{
    CMCON = 7;           				// Comparators disabled
    PORTA = 0b00000000;  				// GPIO low on power up
    PORTB = 0b00000000;  				// GPIO low on power up
    TRISA = 0b00111100;  				// Set I/O directions on PORT A
    TRISB = 0b00001011;  				// Set I/O directions on PORT B
    TMR1ON = 0;          				// Timer 1 disabled
    TMR2ON = 0;			 				// Timer 2 disabled
    IRCF2 = 1; IRCF1 = 1; IRCF0 = 1; 	// Use 8MHz internal oscillator

    CCP1CON = 0;
    ANSEL = 0b00001100;  // Enable ADCs for voltmeter input and 3.3V reference
    ADCON0 = 0b00010000; // Configure ADC modules
    ADCON1 = 0b10100000;
    ADON = 1;            // Turn ADC module on
    INTCON = 0;
    PIE1 = 0;
    PIE2 = 0;
}

void set_outputs(unsigned char out)
{
    seg_g = (out & (1 << 0)) != 0;
    seg_f = (out & (1 << 1)) != 0;
    seg_e = (out & (1 << 2)) != 0;
    seg_d = (out & (1 << 3)) != 0;
    seg_c = (out & (1 << 4)) != 0;
    seg_b = (out & (1 << 5)) != 0;
    seg_a = (out & (1 << 6)) != 0;
    seg_dp = (out & (1 << 7)) != 0;
}

int main(int argc, char** argv) {

    const float Vdd = 3.29f; 			// Reference voltage (measured)
    const float R1 = 47.0f, R2 = 10.0f; // Input potential divider in kilohms
    const float input_attn = R2/(R1+R2);
    unsigned int dac_val;
    unsigned char big_num;
    float v_in_noisy;
    float v_in = 0;
    float v_in_var_noisy;
    float v_in_var = 0;

	// Initialise
    init();

    // Temporary storage for 7-segment display driver values (both init to 0)
    unsigned char SEGS0 = LUT_7SEG[0], SEGS1 = LUT_7SEG[0];

    while (1){

        // Start A/D conversion
        GO_nDONE = 1;

        // Wait until ADC is finished
        while (GO_nDONE){};

        // Retrieve ADC result and convert to volts
        dac_val = (ADRESH << 8) + (unsigned int) ADRESL;
        v_in_noisy = (float) dac_val*Vdd/(1024.0f*input_attn); // 10-bit DAC ==> 1024

		// Also calculate (smoothed) input variance, to detect when user varies voltage
        v_in_var_noisy = (v_in_noisy - v_in)*(v_in_noisy - v_in);
        v_in_var = v_in_var_noisy + 0.8*(v_in_var - v_in_var_noisy);

        // Simple IIR filter to smooth result
        if (v_in_var > 0.09) // Var > 0.3*0.3 ==> User changed input. Track change.
            v_in = v_in_noisy + 0.8*(v_in - v_in_noisy);
        else // Smooth heavily, to prevent display from flickering
            v_in = v_in_noisy + 0.99*(v_in - v_in_noisy);

        // Display to one decimal place if < 10v
        if(v_in >= 10.0f)
        {
            // NOTE: No safety checking of array indices (for speed)
            big_num = v_in/10;
            SEGS1 = LUT_7SEG[big_num]; // tens
            SEGS0 = LUT_7SEG[(unsigned char)v_in-10*big_num]; // units
        }
        else
        {
            big_num = v_in;
            SEGS1 = LUT_7SEG[big_num] | DP_MASK; // units with decimal point
            SEGS0 = LUT_7SEG[(unsigned char) (10.0f*(v_in-big_num))]; // tenths
        }

		// Demultiplex values to the two digits
        select_7seg = 1;
        set_outputs(~SEGS0); // Invert bits for common-anode
        __delay_ms(9); // Approx. equal to delay needed for above calculations
        select_7seg = 0;
        set_outputs(~SEGS1); // Invert bits for common-anode
        // ... now wait 8.7ms for above calculations to complete ...
    }

    return (EXIT_SUCCESS);
}

